// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004                         >>
// << ------------------------------------------- >>
// << <3dmode.cpp>                                >>
// << General 3d mode stuff                       >>
// <<--------------------------------------------->>

// Includes ------------------------------------- >>
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "map.h"
#include "tex_browse.h"

// VARIABLES ----------------------------- >>
int h_line;
int c_sec;

string copy_ftex = "-";
string copy_wtex = "-";

CVAR(Bool, render_determine_hilight, true, 0);

float grav = 0.01f;

// EXTERNAL VARIABLES -------------------- >>
extern float scale;
extern rect_3d_t *selected_rect;
extern int selected_floor, selected_ceil;
extern wallinfo_t *walls;
extern sectinfo_t **sects;
extern Camera camera;
extern Map map;
extern DWORD n_gl_ssectors;
extern bool browser;
extern TextureList tex_list;
extern string msg_text;
extern int msg_count;
extern bool gravity, mix_tex;
extern numlist_t visible_subsectors;
extern gl_vertex *gl_verts;
extern gl_seg *gl_segs;
extern gl_ssect *gl_ssectors;
extern gl_node *gl_nodes;

point3_t cross(point3_t p1, point3_t p2)
{
	point3_t cross_product;	

	cross_product.x = ((p1.y * p2.z) - (p1.z * p2.y));
	cross_product.y = ((p1.z * p2.x) - (p1.x * p2.z));
	cross_product.z = ((p1.x * p2.y) - (p1.y * p2.x));

	return cross_product;
}

float dot_product(point3_t v1, point3_t v2)
{
	return (v1.x * v2.x) + (v1.y * v1.y) + (v1.z * v2.z);
}

/*
float angle(int x1, int y1, int x2, int y2)
{
	float a = atan2(float(y2 - y1), float(x2 - x1));
    return a * 180.0f / 3.1415926535897932384f;
}

float angle_between_vectors(point3_t v1, point3_t v2)
{
	v1.set(v1.normalize());
	v2.set(v2.normalize());

	float m_v1 = 1.0f / v1.magnitude();
	float m_v2 = 1.0f / v2.magnitude();

	v1.set(v1.x * m_v1, v1.y * m_v1, v1.z * m_v1);
	v2.set(v2.x * m_v2, v2.y * m_v2, v2.z * m_v2);

	float theta = acos(v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);

	return theta * 180.0f / 3.1415926535897932384f;
}
*/

float angle_between_vectors(float x1, float y1, float x2, float y2)
{
	float dot = (x1 * x2) + (y1 * y2);
	float len = sqrt(x1 * x1 + y1 * y1) * sqrt(x2 * x2 + y2 * y2);
	float cosin = dot / len;

	return acos(cosin) * 180.0f / 3.1415926535897932384f;
}

// line_intersect: Determines if the camera is looking at a line
// (borrowed from Doom Builder, thanks CodeImp :D)
// ---------------------------------------------------------- >>
float line_intersect(point3_t r1, point3_t r2, float x1, float y1, float x2, float y2)
{
	// Calculate the intersection distance from the ray
	float u_ray = ((x2 - x1) * (r1.y - y1) - (y2 - y1) * (r1.x - x1)) /
				  ((y2 - y1) * (r2.x - r1.x) - (x2 - x1) * (r2.y - r1.y));
	
	// Calculate the intersection distance from the line
	float u_line = ((r2.x - r1.x) * (r1.y - y1) - (r2.y - r1.y) * (r1.x - x1)) /
				   ((y2 - y1) * (r2.x - r1.x) - (x2 - x1) * (r2.y - r1.y));
	
	// Return the distance on the ray if intersecting, or return -1
	if((u_ray >= 0) && (u_ray <= 1024) && (u_line >= 0) && (u_line <= 1)) return u_ray; else return -1;
}

/*
void camera_to_floor()
{
	int sector = determine_sector((int)(camera.position.x / scale), (int)(camera.position.y / scale));

	if (sector != -1)
	{
		float new_height;

		if (map.sectors[sector]->f_height + 40 <= map.sectors[sector]->c_height)
			new_height = (float)(map.sectors[sector]->f_height + 40) * scale;
		else
			new_height = (float)(map.sectors[sector]->c_height - 2) * scale;

		float difference = new_height - camera.position.z;

		if (difference != 0)
		{
			camera.position.z = new_height;
			camera.view.z += difference;
		}
	}
}
*/

void apply_gravity()
{
	int sector = determine_sector((int)(camera.position.x / scale), (int)(camera.position.y / scale));

	if (sector != -1)
	{
		float floor_height = 0.0f;
		bool moved = false;

		if (map.sectors[sector]->f_height + 40 < map.sectors[sector]->c_height)
			floor_height = (float)(map.sectors[sector]->f_height + 40) * scale;
		else
		{
			floor_height = (float)(map.sectors[sector]->c_height - 4) * scale;
			float diff = floor_height - camera.position.z;
			camera.position.z = floor_height;
			camera.view.z += diff;
			grav = 0.01f;
			return;
		}

		if (camera.position.z > floor_height + 0.05f)
		{
			grav = grav * 2;
			camera.position.z -= grav;
			camera.view.z -= grav;
			moved = true;
		}

		if (camera.position.z < floor_height - 0.05f)
		{
			grav = grav * 2;
			camera.position.z += grav;
			camera.view.z += grav;
			moved = true;
		}

		if (camera.position.z > floor_height - grav && camera.position.z < floor_height + grav && moved)
		{
			float diff = floor_height - camera.position.z;
			camera.position.z = floor_height;
			camera.view.z += diff;
			grav = 0.01f;
		}
	}
}

bool determine_seg_side(DWORD s, float x, float y)
{
	// Get x1,y1,x2,y2 of seg
	float x1 = 0.0f;
	float y1 = 0.0f;
	float x2 = 0.0f;
	float y2 = 0.0f;
	long v = gl_segs[s].vertex1;

	if (v & SEG_GLVERTEX)
	{
		x1 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
		y1 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
	}
	else
	{
		x1 = map.verts[v]->x * scale;
		y1 = map.verts[v]->y * scale;
	}

	v = gl_segs[s].vertex2;

	if (v & SEG_GLVERTEX)
	{
		x2 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
		y2 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
	}
	else
	{
		x2 = map.verts[v]->x * scale;
		y2 = map.verts[v]->y * scale;
	}

	if (determine_line_side(x1, y1, x2, y2, x, y))
		return true;
	else
		return false;
}

void determine_3d_hilight()
{
	if (!render_determine_hilight)
		return;

	int count = 0;
	float min_dist = 1024;
	point3_t direction = camera.view - camera.position;

	h_line = -1;
	selected_floor = -1;
	selected_ceil = -1;
	selected_rect = NULL;

	numlist_t sectors_check;

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (walls[l].visible)
		{
			rect_t lrect = map.l_getrect(l);
			float dist = line_intersect(camera.position, camera.view,
				(float)lrect.x1() * scale, (float)lrect.y1() * scale,
				(float)lrect.x2() * scale, (float)lrect.y2() * scale);

			if (dist != -1 && dist < min_dist)
			{

				point3_t hit_point(camera.position.x + (direction.x * dist),
					camera.position.y + (direction.y * dist),
					camera.position.z + (direction.z * dist));

				int side = 0;

				if (determine_line_side(l, camera.position.x / scale, camera.position.y / scale))
					side = 0;
				else
					side = 1;

				/*
				if (side == 0)
					sectors_check.add(map.l_getsector1(l), true);
				else
					sectors_check.add(map.l_getsector2(l), true);
					*/

				// Determine which part of a line we are pointing to (if any)
				if (walls[l].has_mid[side])
				{
					if (hit_point.z >= walls[l].mid[side].br.z && hit_point.z <= walls[l].mid[side].tl.z)
					{
						selected_rect = &walls[l].mid[side];
						h_line = l;
						min_dist = dist;
					}
				}

				if (walls[l].has_up[side])
				{
					if (hit_point.z >= walls[l].up[side].br.z && hit_point.z <= walls[l].up[side].tl.z)
					{
						selected_rect = &walls[l].up[side];
						h_line = l;
						min_dist = dist;
					}
				}

				if (walls[l].has_lo[side])
				{
					if (hit_point.z >= walls[l].lo[side].br.z && hit_point.z <= walls[l].lo[side].tl.z)
					{
						selected_rect = &walls[l].lo[side];
						h_line = l;
						min_dist = dist;
					}
				}
			}
		}
	}

	for (DWORD a = 0; a < visible_subsectors.n_numbers; a++)
	{
		short s = visible_subsectors.numbers[a];

		if (direction.z > 0)
		{
			float height = map.sectors[sects[s]->parent_sector]->c_height * scale;

			if (camera.position.z < height)
			{
				float dist = (height - camera.position.z) / (camera.view.z - camera.position.z);

				if (dist < min_dist)
				{
					point3_t intersection(camera.position.x + (direction.x * dist),
						camera.position.y + (direction.y * dist),
						camera.position.z + (direction.z * dist));
					
					bool in = true;
					DWORD start = gl_ssectors[s].startseg;
					DWORD end = start + gl_ssectors[s].n_segs;
					for (DWORD seg = start; seg < end; seg++)
					{
						if (!determine_seg_side(seg, intersection.x, intersection.y))
						{
							in = false;
							break;
						}
					}

					if (in)
					{
						selected_ceil = sects[s]->parent_sector;
						selected_rect = NULL;
						h_line = -1;
						min_dist = dist;
					}
				}
			}
		}

		if (direction.z < 0)
		{
			float height = map.sectors[sects[s]->parent_sector]->f_height * scale;

			if (camera.position.z > height)
			{
				float dist = (height - camera.position.z) / (camera.view.z - camera.position.z);

				if (dist < min_dist)
				{
					point3_t intersection(camera.position.x + (direction.x * dist),
						camera.position.y + (direction.y * dist),
						camera.position.z + (direction.z * dist));
					
					bool in = true;
					DWORD start = gl_ssectors[s].startseg;
					DWORD end = start + gl_ssectors[s].n_segs;
					for (DWORD seg = start; seg < end; seg++)
					{
						if (!determine_seg_side(seg, intersection.x, intersection.y))
						{
							in = false;
							break;
						}
					}

					if (in)
					{
						selected_floor = sects[s]->parent_sector;
						selected_rect = NULL;
						h_line = -1;
						min_dist = dist;
					}
				}
			}
		}
	}
}

// update_sector: Updates all ssectors/lines that are in a sector
// ----------------------------------------------------------- >>
void update_sector(int sector)
{
	for (DWORD s = 0; s < n_gl_ssectors; s++)
	{
		if (sects[s]->parent_sector == sector)
			setup_ssect(s);
	}

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (map.l_getsector1(l) == sector || map.l_getsector2(l) == sector)
		{
			walls[l].has_mid[0] = false;
			walls[l].has_mid[1] = false;
			walls[l].has_up[0] = false;
			walls[l].has_up[1] = false;
			walls[l].has_lo[0] = false;
			walls[l].has_lo[1] = false;

			if (map.lines[l]->side1 != -1)
			{
				if (map.l_needsmidtex(l, 1))
				{
					setup_wall_rect(l, 1, 0);
					setup_wall_tex(l, 1, 0);
				}

				if (map.l_needsuptex(l, 1))
				{
					setup_wall_rect(l, 1, 1);
					setup_wall_tex(l, 1, 1);
				}

				if (map.l_needslotex(l, 1))
				{
					setup_wall_rect(l, 1, 2);
					setup_wall_tex(l, 1, 2);
				}
			}

			if (map.lines[l]->side2 != -1)
			{
				if (map.l_needsmidtex(l, 2))
				{
					setup_wall_rect(l, 2, 0);
					setup_wall_tex(l, 2, 0);
				}

				if (map.l_needsuptex(l, 2))
				{
					setup_wall_rect(l, 2, 1);
					setup_wall_tex(l, 2, 1);
				}

				if (map.l_needslotex(l, 2))
				{
					setup_wall_rect(l, 2, 2);
					setup_wall_tex(l, 2, 2);
				}
			}
		}
	}

	determine_3d_hilight();
}

// sector_3d_changeheight: Changes floor/ceiling height of hilighted sector
// --------------------------------------------------------------------- >>
void sector_3d_changeheight(bool floor, int amount)
{
	if (h_line == -1)
	{
		if (selected_floor != -1)
		{
			map.s_changeheight(selected_floor, true, amount);
			char temp[6] = "";
			msg_text = "Height: ";
			sprintf(temp, "%d", map.sectors[selected_floor]->f_height);
			msg_text += temp;
			msg_count = 100;
			update_sector(selected_floor);
		}

		if (selected_ceil != -1)
		{
			map.s_changeheight(selected_ceil, false, amount);
			char temp[6] = "";
			msg_text = "Height: ";
			sprintf(temp, "%d", map.sectors[selected_ceil]->c_height);
			msg_text += temp;
			msg_count = 100;
			update_sector(selected_ceil);
		}
	}
	else
	{
		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
		{
			map.s_changeheight(map.l_getsector1(h_line), floor, amount);
			msg_count = 100;

			if (floor)
			{
				char temp[6] = "";
				msg_text = "Floor Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector1(h_line)]->f_height);
				msg_text += temp;
			}
			else
			{
				char temp[6] = "";
				msg_text = "Ceiling Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector1(h_line)]->c_height);
				msg_text += temp;
			}

			update_sector(map.l_getsector1(h_line));
		}
		else
		{
			map.s_changeheight(map.l_getsector2(h_line), floor, amount);
			msg_count = 100;

			if (floor)
			{
				char temp[6] = "";
				msg_text = "Floor Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector2(h_line)]->f_height);
				msg_text += temp;
			}
			else
			{
				char temp[6] = "";
				msg_text = "Ceiling Height: ";
				sprintf(temp, "%d", map.sectors[map.l_getsector2(h_line)]->c_height);
				msg_text += temp;
			}

			update_sector(map.l_getsector2(h_line));
		}
	}
}

void init_3d_browser()
{
	browser = true;

	if (selected_rect == NULL)
	{
		if (selected_floor != -1)
		{
			if (mix_tex)
				setup_browser(true, true);
			else
				setup_browser(false, true);
			
			browser_findtex(map.sectors[selected_floor]->f_tex);
		}

		if (selected_ceil != -1)
		{
			if (mix_tex)
				setup_browser(true, true);
			else
				setup_browser(false, true);

			browser_findtex(map.sectors[selected_ceil]->c_tex);
		}
	}
	else
	{
		if (mix_tex)
				setup_browser(true, true);
			else
				setup_browser(true, false);
		browser_findtex(tex_list.get_name(selected_rect->tex));
	}
}

void copy_3d_texture()
{
	if (selected_rect == NULL)
	{
		if (selected_floor != -1)
		{
			if (mix_tex)
				copy_wtex = map.sectors[selected_floor]->f_tex;

			copy_ftex = map.sectors[selected_floor]->f_tex;
			msg_text = "Copied Texture \"";
			msg_text += copy_ftex;
			msg_text += "\"";
			msg_count = 100;
		}

		if (selected_ceil != -1)
		{
			if (mix_tex)
				copy_wtex = map.sectors[selected_ceil]->c_tex;

			copy_ftex = map.sectors[selected_ceil]->c_tex;
			msg_text = "Copied Texture \"";
			msg_text += copy_ftex;
			msg_text += "\"";
			msg_count = 100;
		}
	}
	else
	{
		if (mix_tex)
			copy_ftex = tex_list.get_name(selected_rect->tex);

		copy_wtex = tex_list.get_name(selected_rect->tex);
		msg_text = "Copied Texture \"";
		msg_text += copy_wtex;
		msg_text += "\"";
		msg_count = 100;
	}
}

void update_selected_rect()
{
	/*
	int side = 0;
	int part = 0;

	if (&walls[h_line].mid[0] == selected_rect)
	{
		side = 1;
		part = 0;
	}
	else if (&walls[h_line].mid[1] == selected_rect)
	{
		side = 2;
		part = 0;
	}
	else if (&walls[h_line].up[0] == selected_rect)
	{
		side = 1;
		part = 1;
	}
	else if (&walls[h_line].up[1] == selected_rect)
	{
		side = 2;
		part = 1;
	}
	else if (&walls[h_line].lo[0] == selected_rect)
	{
		side = 1;
		part = 2;
	}
	else if (&walls[h_line].lo[1] == selected_rect)
	{
		side = 2;
		part = 2;
	}

	setup_wall_rect(h_line, side, part);
	setup_wall_tex(h_line, side, part);
	*/
	int l = h_line;

	walls[l].has_mid[0] = false;
	walls[l].has_mid[1] = false;
	walls[l].has_up[0] = false;
	walls[l].has_up[1] = false;
	walls[l].has_lo[0] = false;
	walls[l].has_lo[1] = false;

	if (map.lines[l]->side1 != -1)
	{
		if (map.l_needsmidtex(l, 1))
		{
			setup_wall_rect(l, 1, 0);
			setup_wall_tex(l, 1, 0);
		}

		if (map.l_needsuptex(l, 1))
		{
			setup_wall_rect(l, 1, 1);
			setup_wall_tex(l, 1, 1);
		}

		if (map.l_needslotex(l, 1))
		{
			setup_wall_rect(l, 1, 2);
			setup_wall_tex(l, 1, 2);
		}
	}

	if (map.lines[l]->side2 != -1)
	{
		if (map.l_needsmidtex(l, 2))
		{
			setup_wall_rect(l, 2, 0);
			setup_wall_tex(l, 2, 0);
		}

		if (map.l_needsuptex(l, 2))
		{
			setup_wall_rect(l, 2, 1);
			setup_wall_tex(l, 2, 1);
		}

		if (map.l_needslotex(l, 2))
		{
			setup_wall_rect(l, 2, 2);
			setup_wall_tex(l, 2, 2);
		}
	}
}

void update_wall_tex(string texture)
{
	int side = 0;
	int part = 0;

	if (&walls[h_line].mid[0] == selected_rect)
	{
		side = 1;
		part = 0;
		map.l_setmidtex(h_line, 1, texture);
	}
	else if (&walls[h_line].mid[1] == selected_rect)
	{
		side = 2;
		part = 0;
		map.l_setmidtex(h_line, 2, texture);
	}
	else if (&walls[h_line].up[0] == selected_rect)
	{
		side = 1;
		part = 1;
		map.l_setuptex(h_line, 1, texture);
	}
	else if (&walls[h_line].up[1] == selected_rect)
	{
		side = 2;
		part = 1;
		map.l_setuptex(h_line, 2, texture);
	}
	else if (&walls[h_line].lo[0] == selected_rect)
	{
		side = 1;
		part = 2;
		map.l_setlotex(h_line, 1, texture);
	}
	else if (&walls[h_line].lo[1] == selected_rect)
	{
		side = 2;
		part = 2;
		map.l_setlotex(h_line, 2, texture);
	}

	setup_wall_rect(h_line, side, part);
	setup_wall_tex(h_line, side, part);
}

void update_room_tex(string texture)
{
	int side = 0;
	int part = 0;
	string old_tex = tex_list.get_name(selected_rect->tex);

	if (&walls[h_line].mid[0] == selected_rect)
	{
		side = 1;
		part = 0;
		map.l_setmidtex(h_line, 1, texture);
	}
	else if (&walls[h_line].mid[1] == selected_rect)
	{
		side = 2;
		part = 0;
		map.l_setmidtex(h_line, 2, texture);
	}
	else if (&walls[h_line].up[0] == selected_rect)
	{
		side = 1;
		part = 1;
		map.l_setuptex(h_line, 1, texture);
	}
	else if (&walls[h_line].up[1] == selected_rect)
	{
		side = 2;
		part = 1;
		map.l_setuptex(h_line, 2, texture);
	}
	else if (&walls[h_line].lo[0] == selected_rect)
	{
		side = 1;
		part = 2;
		map.l_setlotex(h_line, 1, texture);
	}
	else if (&walls[h_line].lo[1] == selected_rect)
	{
		side = 2;
		part = 2;
		map.l_setlotex(h_line, 2, texture);
	}

	setup_wall_rect(h_line, side, part);
	setup_wall_tex(h_line, side, part);
}

void update_sector_flats(int sector)
{
	for (DWORD s = 0; s < n_gl_ssectors; s++)
	{
		if (sects[s]->parent_sector == sector)
			setup_ssect(s);
	}
}

void paste_3d_texture(bool paint_room)
{
	if (h_line == -1 && copy_ftex != "-")
	{
		if (selected_floor != -1)
		{
			map.sectors[selected_floor]->f_tex = copy_ftex;
			update_sector(selected_floor);
		}
		if (selected_ceil != -1)
		{
			map.sectors[selected_ceil]->c_tex = copy_ftex;
			update_sector(selected_ceil);
		}
	}
	else if (h_line != -1)
	{
		if (paint_room)
			update_room_tex(copy_wtex);
		else
			update_wall_tex(copy_wtex);
	}
}

void set_3d_texture()
{
	if (h_line == -1)
	{
		if (selected_floor != -1)
		{
			map.sectors[selected_floor]->f_tex = browser_gettex();
			update_sector(selected_floor);
		}
		if (selected_ceil != -1)
		{
			map.sectors[selected_ceil]->c_tex = browser_gettex();
			update_sector(selected_ceil);
		}
	}
	else
		update_wall_tex(browser_gettex());
}

void change_3d_sector_light(int amount)
{
	int sector = -1;

	if (h_line == -1)
	{
		if (selected_floor != -1)
			sector = selected_floor;
		if (selected_ceil != -1)
			sector = selected_ceil;
	}
	else if (h_line != -1)
	{
		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			sector = map.l_getsector1(h_line);
		else
			sector = map.l_getsector2(h_line);
	}

	if (sector == -1)
		return;
	else
	{
		int light = map.sectors[sector]->light + amount;

		if (light == 239)	// Original light was 255, so only subtract 15 to keep 
			light = 240;	// the light level divisible by 16
		if (light > 255)
			light = 255;
		if (light < 0)
			light = 0;

		map.sectors[sector]->light = light;
		update_sector_flats(sector);
		char temp[5] = "";
		sprintf(temp, "%d", light);
		msg_text = "Light Level: ";
		msg_text += temp;
		msg_count = 100;
	}
}

void change_3d_texture_offset(int x_amount, int y_amount)
{
	if (h_line != -1)
	{
		int side = -1;

		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			side = map.lines[h_line]->side1;
		else
			side = map.lines[h_line]->side2;

		if (side != -1)
		{
			map.sides[side]->x_offset += x_amount;
			map.sides[side]->y_offset += y_amount;

			if (x_amount != 0)
			{
				char temp[8];
				sprintf(temp, "%d", map.sides[side]->x_offset);
				msg_text = "X Offset: ";
				msg_text += temp;
				msg_count = 100;
			}

			if (y_amount != 0)
			{
				char temp[8];
				sprintf(temp, "%d", map.sides[side]->y_offset);
				msg_text = "Y Offset: ";
				msg_text += temp;
				msg_count = 100;
			}
		}

		update_selected_rect();
	}
}

void toggle_3d_texpeg(bool upper)
{
	if (h_line == -1)
		return;

	if (upper)
	{
		if (map.lines[h_line]->toggle_flag(LINE_UPPERUNPEGGED))
			msg_text = "Upper Unpegged: On";
		else
			msg_text = "Upper Unpegged: Off";

		msg_count = 100;
	}
	else
	{
		if (map.lines[h_line]->toggle_flag(LINE_LOWERUNPEGGED))
			msg_text = "Lower Unpegged: On";
		else
			msg_text = "Lower Unpegged: Off";

		msg_count = 100;
	}

	update_selected_rect();
}

void delete_3d_texture()
{
	if (h_line != -1)
		update_wall_tex("-");
}

void reset_3d_tex_offsets()
{
	if (h_line != -1)
	{
		int side = -1;

		if (determine_line_side(h_line, camera.position.x / scale, camera.position.y / scale))
			side = map.lines[h_line]->side1;
		else
			side = map.lines[h_line]->side2;

		if (side != -1)
		{
			map.sides[side]->x_offset = 0;
			map.sides[side]->y_offset = 0;
		}

		update_selected_rect();
	}
}
