#include "main.h"
#include "draw.h"

float	console_height = 0.0f;		// Current height of console
float	console_maxheight = 0.4f;	// Maximum height of console (how muc hof the screen it takes up)
bool	console = false;			// Console active?

stringlist_t	console_history;	// History of commands (for retrieving previous commands)
stringlist_t	console_log;		// Log of console (for display)
string			console_command;	// Current command line
int				console_hcmd = -1;	// Current previous command retrieved

rgba_t	console_text_command(255, 255, 255, 255);
rgba_t	console_text_history(160, 170, 180, 255);
rgba_t	console_background(0, 50, 100, 180);

EXTERN_CVAR(Int, vid_width)
EXTERN_CVAR(Int, vid_height)

void draw_console()
{
	if (console)
	{
		if (console_height < console_maxheight)
			console_height += 0.02f;
	}
	else
		console_height -= 0.02f;

	int height = (int)(vid_height * console_height);

	// Draw console background & line
	rgba_t colour(console_background.r, console_background.g, console_background.b,
					(console_height / console_maxheight) * console_background.a);
	draw_rect(rect_t(0, 0, vid_width, height, RECT_TOPLEFT), colour, true);
	colour.set(255, 255, 255, (console_height / console_maxheight) * 255);
	draw_line(rect_t(0, height, vid_width, 0, RECT_TOPLEFT), colour, true);

	// Draw console command line
	colour.set(console_text_command.r, console_text_command.g, console_text_command.b,
					(console_height / console_maxheight) * console_text_command.a);
	draw_text(8, height - 16, colour, 0, ") %s", console_command.c_str());
	draw_text(24 + (console_command.length() * 8), height - 16, colour, 0, "_");

	// Draw SLADE version
	colour.set(console_text_history.r, console_text_history.g, console_text_history.b,
					(console_height / console_maxheight) * console_text_history.a);
	draw_text(vid_width - 8, height - 16, colour, 2, "SLADE v0.4");

	// Draw console log
	bool done = false;
	int y = height - 26;
	int line = console_log.n_strings;
	while (!done)
	{
		if (line == 0)
			done = true;
		else if (y < 0)
			done = true;
		else
			draw_text(8, y, colour, 0, console_log.strings[line - 1]->c_str());

		line--;
		y -= 10;
	}
}

void console_prevcommand()
{
	if (console_hcmd != -1)
		console_hcmd--;

	if (console_hcmd == -1)
	{
		if (console_history.n_strings > 0)
			console_hcmd = console_history.n_strings - 1;
	}

	if (console_hcmd != -1)
		console_command = *console_history.strings[console_hcmd];
}

void console_nextcommand()
{
	if (console_hcmd == -1)
		return;

	if (console_hcmd != -1)
		console_hcmd++;

	if (console_hcmd >= console_history.n_strings)
	{
		console_command = "";
		console_hcmd = -1;
	}
	else
		console_command = *console_history.strings[console_hcmd];
}

void console_parsecommand()
{
	if (console_command == "")
		return;

	// Add command to history
	console_history.add(console_command);

	bool parsed = false;
	MemReader mr;
	mr.open_string(console_command, 0, 0);

	string token = mr.get_token();

	// Check for cvar
	CVar* cvar = get_cvar(token);
	if (cvar)
	{
		bool changed = false;

		// Check if we want to change the value
		if (mr.peek_token() != "!END")
		{
			if (cvar->type == CVAR_INTEGER)
				*((CIntCVar *)cvar) = mr.get_integer();

			if (cvar->type == CVAR_BOOLEAN)
				*((CBoolCVar *)cvar) = !!(mr.get_integer());

			if (cvar->type == CVAR_FLOAT)
				*((CFloatCVar *)cvar) = mr.get_float();

			if (cvar->type == CVAR_STRING)
				*((CStringCVar *)cvar) = mr.get_token();

			changed = true;
		}

		// Display the value
		string msg = "- \"";
		msg += cvar->name;

		if (changed)
			msg += "\" set to ";
		else
			msg += "\" is ";

		char val[16] = "";

		if (cvar->type == CVAR_INTEGER)
			sprintf(val, "\"%d\"", cvar->GetValue().Int);

		if (cvar->type == CVAR_STRING)
			sprintf(val, "\"%s\"", ((CStringCVar *)cvar)->value.c_str());

		if (cvar->type == CVAR_BOOLEAN)
			sprintf(val, "\"%d\"", cvar->GetValue().Bool);

		if (cvar->type == CVAR_FLOAT)
			sprintf(val, "\"%1.2f\"", cvar->GetValue().Float);

		console_log.add(msg + val);

		parsed = true;
	}

	// "cvarlist" command
	if (token == "cvarlist")
	{
		console_log.add("- All CVars:");

		stringlist_t l_cvars;
		get_cvar_list(&l_cvars);

		for (DWORD s = 0; s < l_cvars.n_strings; s++)
		{
			char temp[128] = "";
			sprintf(temp, "\"%s\"", l_cvars.strings[s]->c_str());
			console_log.add(temp);
		}

		parsed = true;
	}

	// "cmdlist" command
	if (token == "cmdlist")
	{
		console_log.add("- Valid Commands:");
		console_log.add("\"cmdlist\"");
		console_log.add("\"cvarlist\"");

		parsed = true;
	}

	// Unknown command
	if (!parsed)
	{
		char temp[276] = "";
		sprintf(temp, "- Unknown Command \"%s\"", console_command.c_str());
		console_log.add(temp);
	}

	// Finish up
	console_command = "";
	console_hcmd = -1;
}
