// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004                  >>
// << ------------------------------------ >>
// << main.cpp - All the main stuff        >>
// << ------------------------------------ >>

// INCLUDES ------------------------------ >>
#include "main.h"
#include "gui.h"
#include "menu.h"
#include "input.h"
#include "keybind.h"
#include "map.h"
#include "textures.h"
#include "struct_3d.h"
#include "camera.h"
#include "3dmode.h"

// VARIABLES ----------------------------- >>
SDL_Surface	*screen;	// The main sreen surface

Uint32	next_time;

// Video settings
CVAR(Int, vid_width, 1024, CVAR_SAVE)
CVAR(Int, vid_height, 768, CVAR_SAVE)
CVAR(Int, vid_bitdepth, 32, CVAR_SAVE)
CVAR(Int, vid_fullscreen, 0, CVAR_SAVE)

CVAR(String, colour_scheme, "Blue", CVAR_SAVE)
CVAR(String, game_config, "doom2", CVAR_SAVE)

CVAR(Float, mouse_sensitivity, 0.5f, CVAR_SAVE)

bool running = true;
bool mode_3d = false;

bool mouse1 = false;
bool mouse2 = false;
bool mouse_wheelup = false;
bool mouse_wheeldown = false;
int mouse1_c = 0;
int mouse2_c = 0;
bool items_moving = false;

string game = "-";	// Open game configuration

point2_t	mouse;	// Mouse pointer position

stringlist_t	search_paths;

// EXTERNAL VARIABLES -------------------- >>
extern Map map;
extern WadList wads;
extern bool menu, inbox, browser, message, line_draw, console;
extern Camera camera;

extern string def_ftex, def_ctex, def_midtex, def_uptex, def_lotex;
extern short def_fheight, def_cheight, def_light;
extern int zoom;
extern rect_t sel_box;
extern stringlist_t popup_message;
extern int key_3d_rep;
extern bool mix_tex;

// message_box: Pops up an OS message box
// ----------------------------------- >>
void message_box(string message)
{
#ifdef WIN32
	MessageBox(NULL, message.c_str(), "SLADE", MB_OK);
#endif
}

// bool2yesno: Returns either 'yes' or 'no' depending on supplied boolean value
// ------------------------------------------------------------------------- >>
string bool2yesno(bool value)
{
	if (value)
		return "Yes";
	else
		return "No";
}

// str_upper: Returns a string in uppercase
// ------------------------------------- >>
string str_upper(string str)
{
	DWORD len = 0;
	char* ustr = NULL;

	for (DWORD c = 0; c < str.length(); c++)
		len++;

	ustr = (char *)malloc(len + 1);

	for (c = 0; c < str.length(); c++)
		ustr[c] = str[c];

	ustr[len] = 0;

	string ret = strupr(ustr);
	free(ustr);
	return ret;
}

// read_line: Reads a line of text from a file
// ---------------------------------------- >>
void read_line(FILE *f, char *string)
{
	do
	{
		fgets(string, 255, f);
	} while ((string[0] == '/') || (string[0] == '\n'));
}

void read_types(MemReader *mr, bool things, bool lines, bool sectors, bool args)
{
	string token = mr->get_token();
	while (token != "!END")
	{
		// Thing types
		if (things && token == "thing_types")
		{
			mr->check_token("{");

			token = mr->get_token();
			while (token != "}")
			{
				if (token == "group")
					parse_thing_group(mr);

				if (token == "import_types")
				{
					MemReader mr2;
					mr2.open_file("games/" + mr->get_token(), 0, 0);
					read_types(&mr2, true, false, false, false);
				}

				token = mr->get_token();
			}
		}

		// Line types
		if (lines && token == "line_types")
		{
			mr->check_token("{");

			token = mr->get_token();
			while (token != "}")
			{
				if (token == "group")
					parse_line_group(mr);

				if (token == "import_types")
				{
					MemReader mr2;
					mr2.open_file("games/" + mr->get_token(), 0, 0);
					read_types(&mr2, false, true, false, false);
				}

				token = mr->get_token();
			}
		}

		// Sector types
		if (sectors && token == "sector_types")
		{
			mr->check_token("{");

			token = mr->get_token();
			while (token != "}")
			{
				if (token == "type")
					parse_sectortype(mr);

				if (token == "import_types")
				{
					MemReader mr2;
					mr2.open_file("games/" + mr->get_token(), 0, 0);
					read_types(&mr2, false, false, true, false);
				}

				token = mr->get_token();
			}
		}

		// Arg types
		if (args && token == "arg_types")
		{
			mr->check_token("{");

			token = mr->get_token();
			while (token != "}")
			{
				if (token == "type")
					parse_argtype(mr);

				if (token == "import_types")
				{
					MemReader mr2;
					mr2.open_file("games/" + mr->get_token(), 0, 0);
					read_types(&mr2, false, false, false, true);
				}

				token = mr->get_token();
			}
		}

		token = mr->get_token();
	}
}

bool read_game_config(string path)
{
	MemReader mr;
	if (!mr.open_file(path, 0, 0))
	{
		message_box("Game configuration file \"" + path + "\" not found!\nPlease check slade.cfg.");
		return false;
	}

	// Check for game section
	if (mr.get_token() != "game")
	{
		message_box("Invalid game configuration file, no 'game' section found! (must be first)");
		return false;
	}

	// Read game name
	game = mr.get_token();

	// Check for opening brace
	if (!mr.check_token("{"))
		return false;

	// Read game info
	string token = mr.get_token();
	while (token != "}")
	{
		/*
		if (token == "map_format")
		{
			int format = mr.get_integer();
			
			if (format >= 1)
				map.hexen = true;

			if (format == 2)
			{
				map.zdoom = true;
				mix_tex = true;
			}
		}
		*/

		if (token == "boom")
			map.boom = mr.get_bool();

		if (token == "hexen")
			map.hexen = mr.get_bool();

		if (token == "zdoom")
		{
			map.zdoom = mr.get_bool();
			mix_tex = true;
		}

		if (token == "iwad")
		{
			string iwad = mr.get_token();
			if (!wads.open_iwad(iwad))
			{
				message_box("Game IWAD \"" + iwad + "\" not found!\nCorrect the path in \"" + path + "\"");
				return false;
			}
		}

		token = mr.get_token();
	}

	// Read defaults
	if (mr.get_token() == "defaults")
	{
		mr.check_token("{");
		token = mr.get_token();

		while (token != "}")
		{
			if (token == "floor_tex")
				def_ftex = mr.get_token();

			if (token == "ceil_tex")
				def_ctex = mr.get_token();

			if (token == "upper_tex")
				def_uptex = mr.get_token();

			if (token == "lower_tex")
				def_lotex = mr.get_token();

			if (token == "middle_tex")
				def_midtex = mr.get_token();

			if (token == "floor_height")
				def_fheight = mr.get_integer();

			if (token == "ceil_height")
				def_cheight = mr.get_integer();

			if (token == "light")
				def_light = mr.get_integer();

			token = mr.get_token();
		}
	}

	//printf("Read game info ok\n");

	// Read thing/line/sector types
	read_types(&mr, true, true, true, true);

	return true;
}

void load_search_paths(MemReader *mr)
{
	if (!mr->check_token("{"))
		return;

	string token = mr->get_token();

	while (token != "}" && token != "!END")
	{
		search_paths.add(token);
		token = mr->get_token();
	}
}

void save_search_paths(FILE *fp)
{
	fprintf(fp, "paths\n{\n");

	for (DWORD s = 0; s < search_paths.n_strings; s++)
		fprintf(fp, "\t%s\n", search_paths.strings[s]->c_str());

	fprintf(fp, "}\n\n");
}

void load_main_config()
{
	MemReader mr;
	mr.open_file("slade.cfg", 0, 0);
	string token = mr.get_token();

	while (token != "!END")
	{
		if (token == "cvars")
			load_cvars(&mr);

		if (token == "paths")
			load_search_paths(&mr);

		token = mr.get_token();
	}
}

void save_main_config()
{
	FILE *fp = fopen("slade.cfg", "wt");
	save_search_paths(fp);
	save_cvars(fp);
	fclose(fp);
}

// init_opengl: Initialises OpenGL
// ---------------------------- >>
bool init_opengl()
{
	if (mode_3d)
	{
		GLfloat fogColor[4]= {0.0f, 0.0f, 0.0f, 1.0f};

		glViewport(0, 0, vid_width, vid_height);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClearDepth(1.0);
		glShadeModel(GL_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_CULL_FACE);
		glEnable(GL_TEXTURE_2D);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LEQUAL);
		glDepthRange(0, 1);

		glFogi(GL_FOG_MODE, GL_LINEAR);
		glFogfv(GL_FOG_COLOR, fogColor);
		glFogf(GL_FOG_DENSITY, 1.0f);
		glHint(GL_FOG_HINT, GL_NICEST);
		glFogf(GL_FOG_START, 0.0f);
		glFogf(GL_FOG_END, 40.0f);
		glEnable(GL_FOG);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		gluPerspective(60.0f, (GLfloat)320/(GLfloat)200, 0.1f, 100.0f);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}
	else
	{
		glViewport(0, 0, vid_width, vid_height);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClearDepth(1.0);
		glShadeModel(GL_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glCullFace(GL_NONE);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glOrtho(0.0f, vid_width, vid_height, 0.0f, -1.0f, 1.0f);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}

	return true;
}

// init: Initialises the SLADE editor
// ------------------------------- >>
bool init()
{
	// Init SDL video
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		printf("Error Initialising SDL Video: %s", SDL_GetError());
		return false;
	}

	// Setup SDL stuff
	atexit(SDL_Quit);
	SDL_WM_SetCaption("SLADE", "");

	if (vid_fullscreen)
		screen = SDL_SetVideoMode(vid_width, vid_height, vid_bitdepth, SDL_OPENGL|SDL_OPENGLBLIT|SDL_FULLSCREEN);
	else
		screen = SDL_SetVideoMode(vid_width, vid_height, vid_bitdepth, SDL_OPENGL|SDL_OPENGLBLIT|SDL_RESIZABLE);

	SDL_EnableKeyRepeat(50, 50);
	SDL_ShowCursor(false);

	// Check if screen was created
	if (!screen)
	{
		printf("Error creating SDL screen: %s\n", SDL_GetError());
		return false;
	}

	if (!init_opengl())
		return false;

	ilInit();
	load_editor_textures();

	read_palette(wads.get_wad_with_lump("PLAYPAL"));

	if (map.zdoom)
		load_tx_textures();

	gui_drawsplash("Loading flats...");
	load_flats();
	gui_drawsplash("Loading textures...");
	load_textures();
	gui_drawsplash("Loading sprites...");
	load_sprites();
	update_tdef_textures();

	update_map();
	update_grid();

	return true;
}

// end: Does stuff on exiting SLADE
// ----------------------------- >>
void end()
{
	printf("Thankyou for using SLADE.\n");
	save_main_config();

	remove("sladetemp.wad");
	remove("sladetemp.gwa");

	exit(0);
}

// main_loop: The main program loop, returns true on exit
// --------------------------------------------------- >>
bool main_loop()
{
	bool updatescreen = false;

	SDL_Event event;

	while (SDL_PollEvent(&event))
	{
		if (event.type == SDL_QUIT)
			return true;

		if (event.type == SDL_KEYDOWN)
		{
			//printf("Pressed key %d\n", event.key.keysym.sym);

			if (event.key.keysym.sym == SDLK_F11)
				return true;

			if (!mode_3d)
			{
				if (browser)
				{
					if (!keys_browser())
						browser = false;
				}
				else if (message)
				{
					if (event.key.keysym.sym == SDLK_RETURN
						|| event.key.keysym.sym == SDLK_ESCAPE)
					{
						message = false;
						popup_message.clear();
					}
				}
				else if (inbox)
					keys_inputbox();
				else if (menu)
				{
					if (!keys_menu())
						return true;
				}
				else if (line_draw)
					keys_linedraw();
				else if (console)
					keys_console();
				else
					keys_edit();
			}
			else
			{
				if (browser)
				{
					if (!keys_browser())
						browser = false;
				}
				else if (console)
					keys_console();
			}

			updatescreen = true;
		}

		if (event.type == SDL_KEYUP)
		{
			if (mode_3d)
				key_3d_rep = 0;
		}

		if (event.type == SDL_MOUSEMOTION && !message && !inbox)
		{
			mouse.set(event.motion.x, event.motion.y);
			
			if (menu)
				check_mouse_menu();
			else
			{
				if (mouse2)
				{
					move_items();
					items_moving = true;
				}

				if (mode_3d && event.motion.x != vid_width / 2 && event.motion.y != vid_height / 2)
				{
					if (!browser)
					{
						float angle_x = -event.motion.xrel * mouse_sensitivity * 0.001f;
						float angle_y = -event.motion.yrel * mouse_sensitivity * 0.001f;

						point3_t axis = cross(camera.view - camera.position, camera.up_vector);
						axis = axis.normalize();

						camera.RotateView(angle_x, 0, 0, 1);
						camera.RotateView(angle_y, axis.x, axis.y, axis.z);

						determine_3d_hilight();
						
						if (!vid_fullscreen)
							SDL_WarpMouse(vid_width / 2, vid_height / 2);
					}
				}

				if (sel_box.x1() != -1)
					sel_box.set(sel_box.x1(), sel_box.y1(), mouse.x, mouse.y);
			}

			updatescreen = true;
		}

		// Unfortunately, when SDL_SetVideoMode is called, the OpenGL context is lost.
		// That means all textures are lost also, so all that can be done is reload all
		// the textures upon a screen resize :/
		if (event.type == SDL_VIDEORESIZE)
		{
			vid_width = event.resize.w;
			vid_height = event.resize.h;
			init();

			updatescreen = true;
		}

		if (event.type == SDL_MOUSEBUTTONDOWN)
		{
			if (event.button.button == SDL_BUTTON_LEFT)
			{
				mouse1  = true;
				mouse1_c = 10;
				
				if (mode_3d)
				{
					if (SDL_GetModState() & KMOD_CTRL)
						copy_3d_texture();
					else
						init_3d_browser();
				}
				else 
				{
					if (menu)
					{
						if (check_mouse_menu())
							menu_action();
						else
							menu = false;
					}
					else
					{
						if (line_draw)
							line_drawpoint();
						else
						{
							if (mouse.y < 16)
								menubar_click(mouse.x);
							else
							{
								if (SDL_GetModState() & KMOD_SHIFT)
									sel_box.set(mouse.x, mouse.y, mouse.x, mouse.y);
								else
									select_item();
							}
						}
					}
				}
			}

			if (event.button.button == SDL_BUTTON_RIGHT)
			{
				mouse2 = true;
				mouse2_c = 10;

				if (mode_3d)
					paste_3d_texture(!!(SDL_GetModState() & KMOD_SHIFT));
				else
				{
					if (line_draw)
						line_undrawpoint();
					else
						add_move_items();
				}
			}

			if (event.button.button == SDL_BUTTON_MIDDLE)
			{
				if (mode_3d)
					copy_3d_texture();
			}

			if (event.button.button == SDL_BUTTON_WHEELUP)
			{
				zoom *= 2;

				if (zoom > 128)
					zoom = 128;

				update_map();
				update_grid();
			}

			if (event.button.button == SDL_BUTTON_WHEELDOWN)
			{
				zoom /= 2;

				if (zoom < 1)
					zoom = 1;

				update_map();
				update_grid();
			}

			updatescreen = true;
		}

		if (event.type == SDL_MOUSEBUTTONUP)
		{
			if (event.button.button == SDL_BUTTON_RIGHT)
			{
				if (!mode_3d)
				{
					if (!line_draw)
					{
						clear_move_items();

						if (!items_moving)
							edit_item();
					}
				}

				mouse2 = false;
				items_moving = false;
			}

			if (event.button.button == SDL_BUTTON_LEFT)
				mouse1 = false;

			if (sel_box.x1() != -1)
			{
				select_items_box();
				sel_box.set(-1, -1, -1, -1);
			}

			updatescreen = true;
		}
	}

	if (mouse1_c > 0)
		mouse1_c--;

	if (mouse2_c > 0)
		mouse2_c--;

	if (mode_3d)
	{
		if (!browser)
		{
			if (!console)
				keys_3d();

			camera.Update();
		}

		if (mode_3d)
			render_3dmode();
	}
	else
	{
		if (updatescreen)
		{
			if (!mouse2 && !menu && !line_draw)
				get_hilight_item();

			draw_2d_screen();
		}
	}

	return false;
}

// time_left: Gets the time remaining until the next main loop run
// ------------------------------------------------------------ >>
Uint32 time_left()
{
	Uint32 now;

	now = SDL_GetTicks();
	if(next_time <= now)
		return 0;
	else
		return next_time - now;
}

// main: The main function
// -------------------- >>
int main(int argc, char *argv[])
{
	// Load key binds
	load_binds();

	// Load cvars
	//load_cvars("slade.cfg");

	// Load config
	load_main_config();

	// Load colour schemes
	if (!load_colour_configs("config/colours.cfg"))
	{
		message_box("config/colours.cfg not found!");
		end();
		return 0;
	}

	// Pre-init menu
	menu_pre_init();

	// Load game configuration
	string gc = game_config;

	if (!read_game_config("games/" + gc + ".cfg"))
	{
		end();
		return 0;
	}

	set_colour_config(colour_scheme);

	// Init menu
	menu_init();

	// Display welcome message
	text_welcome();

	while (1)
	{
		text_ui();

		if (!running)
			break;

		init();

		while (!main_loop())
		{
			SDL_Delay(time_left());
			next_time += 15;
		}

		SDL_Quit();
	}

	end();
	return 0;
}
