// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004                  >>
// << ------------------------------------ >>
// << render.cpp: 3d rendering functions   >>
// << ------------------------------------ >>

#pragma warning (disable: 4244)
#pragma warning (disable: 4018)

// INCLUDES ------------------------------ >>
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "map.h"
#include "draw.h"
#include "gui.h"
#include "tex_browse.h"
#include "console.h"
#include "3dmode.h"
#include "misc.h"

// STRUCTS ------------------------------- >>
struct t_dist
{
	short thing;
	double distance;
};

// VARIABLES ----------------------------- >>
Camera		camera;
wallinfo_t	*walls = NULL;
sectinfo_t	**sects = NULL;
float		scale = 0.01f;
GLuint		sky_tex_id = 0;
string		skytex = "SKY1";

sprite_3d_t	*sprites3d = NULL;
numlist_t	sprite_order;

gl_vertex	*gl_verts = NULL;
gl_seg		*gl_segs = NULL;
gl_ssect	*gl_ssectors = NULL;
gl_node		*gl_nodes = NULL;
DWORD		n_gl_verts = 0;
DWORD		n_gl_segs = 0;
DWORD		n_gl_ssectors = 0;
DWORD		n_gl_nodes = 0;

rgba_t		col_3d_crosshair(100, 180, 255, 180);
rgba_t		col_3d_hilight(0, 140, 220, 50);
rgba_t		col_3d_hilight_line(0, 200, 255, 180);

string		msg_text;
int			msg_count = 0;

rect_3d_t	*selected_rect = NULL;
int			selected_floor = -1;
int			selected_ceil = -1;

Uint8		vis_buffer[3600];
numlist_t	visible_subsectors;

int			ltest = 255;

CVAR(Bool, render_hilight, true, CVAR_SAVE);
CVAR(Bool, render_things, false, CVAR_SAVE);
CVAR(Bool, render_fullbright, false, CVAR_SAVE);
CVAR(Bool, render_sky_fullbright, true, CVAR_SAVE);
CVAR(Bool, render_debuginfo, false, 0);
CVAR(Bool, render_bsp, true, 0);
CVAR(Bool, render_ssect_lines, false, 0);

// EXTERNAL VARIABLES -------------------- >>
extern Map map;
extern TextureList tex_list;
extern int h_line, c_sec;
extern bool browser, console, gravity;
extern float console_height;

EXTERN_CVAR(Int, vid_width)
EXTERN_CVAR(Int, vid_height)

void build_nodes()
{
	FILE *fp;
	int current_lump = 0;
	int unit_size = 0;
	int unit_count = 0;

	// Save map and build nodes
	Wad tempwad;
	tempwad.path = "sladetemp.wad";
	map.add_to_wad(&tempwad);
	tempwad.save(false);
	system("glbsp -v3 sladetemp.wad");
	remove("sladetemp.wad.bak");

	// Open GWA file
	Wad wad;
	
	if (!wad.open("sladetemp.gwa"))
		printf("Failed to build GL nodes!\n");

	// << ---------------------- >>
	// << -- Load GL vertices -- >>
	// << ---------------------- >>
	current_lump = wad.get_lump_index("GL_VERT");

	if (current_lump == -1)
	{
		print(true, "Gwa has no GL_VERT lump.\n");
		return;
	}
	
	// Setup sizes
	unit_size = sizeof(gl_vertex);
	unit_count = (wad.directory[current_lump]->Size() - 4) / unit_size;
	n_gl_verts = unit_count;
	
	// Setup array
	gl_verts = (gl_vertex *)realloc(gl_verts, unit_size * unit_count);
	
	// Read vertices from file
	int temp;

	fp = fopen(wad.path.c_str(), "rb");
	fseek(fp, wad.directory[current_lump]->Offset() + 4, SEEK_SET);
	for (DWORD v = 0; v < n_gl_verts; v++)
	{
		fread(&temp, 4, 1, fp);
		gl_verts[v].x = (temp >> 16);
		fread(&temp, 4, 1, fp);
		gl_verts[v].y = (temp >> 16);
	}

	// << ------------------ >>
	// << -- Load GL segs -- >>
	// << ------------------ >>
	current_lump = wad.get_lump_index("GL_SEGS");

	if (current_lump == -1)
	{
		print(true, "Gwa has no GL_SEGS lump.\n");
		return;
	}
	
	// Setup sizes
	unit_size = sizeof(gl_seg);
	unit_count = (wad.directory[current_lump]->Size() - 4) / unit_size;
	n_gl_segs = unit_count;
	
	// Setup array
	gl_segs = (gl_seg *)realloc(gl_segs, unit_size * unit_count);
	
	fseek(fp, wad.directory[current_lump]->Offset() + 4, SEEK_SET);
	fread(gl_segs, unit_size, unit_count, fp);

	//for (int o = 0; o < unit_count; o++)
		//print(true, "%d %d\n", gl_segs[o].vertex1, gl_segs[o].vertex2);

	// << ---------------------- >>
	// << -- Load GL ssectors -- >>
	// << ---------------------- >>
	current_lump = wad.get_lump_index("GL_SSECT");

	if (current_lump == -1)
	{
		print(true, "Gwa has no GL_SSECT lump.\n");
		return;
	}
	
	// Setup sizes
	unit_size = sizeof(gl_ssect);
	unit_count = (wad.directory[current_lump]->Size() - 4) / unit_size;
	n_gl_ssectors = unit_count;
	
	// Setup array
	gl_ssectors = (gl_ssect *)realloc(gl_ssectors, unit_size * unit_count);
	
	fseek(fp, wad.directory[current_lump]->Offset() + 4, SEEK_SET);
	fread(gl_ssectors, unit_size, unit_count, fp);

	// << ------------------- >>
	// << -- Load GL nodes -- >>
	// << ------------------- >>
	current_lump = wad.get_lump_index("GL_NODES");

	if (current_lump == -1)
	{
		print(true, "Gwa has no GL_NODES lump.\n");
		return;
	}
	
	// Setup sizes
	unit_size = sizeof(gl_node);
	unit_count = (wad.directory[current_lump]->Size()) / unit_size;
	n_gl_nodes = unit_count;
	
	// Setup array
	gl_nodes = (gl_node *)realloc(gl_nodes, unit_size * unit_count);
	
	fseek(fp, wad.directory[current_lump]->Offset(), SEEK_SET);
	for (int n = 0; n < n_gl_nodes; n++)
	{
		fread(&gl_nodes[n].x, 2, 1, fp);
		fread(&gl_nodes[n].y, 2, 1, fp);
		fread(&gl_nodes[n].dx, 2, 1, fp);
		fread(&gl_nodes[n].dy, 2, 1, fp);
		fread(gl_nodes[n].right_bbox, 2, 4, fp);
		fread(gl_nodes[n].left_bbox, 2, 4, fp);
		fread(&gl_nodes[n].right_child, 2, 1, fp);
		fread(&gl_nodes[n].left_child, 2, 1, fp);
	}

	fclose(fp);
}

void set_light(short light)
{
	if (render_fullbright)
		light = 255;

	glColor3f(light/255.0f, light/255.0f, light/255.0f);
	glFogf(GL_FOG_END, ((light/255.0f) * 30.0f) + 1.0f);
}

void render_flat_hilight()
{
	float height;

	if (selected_floor != -1)
		height = map.sectors[selected_floor]->f_height * scale;
	else if (selected_ceil != -1)
		height = map.sectors[selected_ceil]->c_height * scale;
	else
		return;

	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	set_colour(col_3d_hilight_line);
	glLineWidth(2.0f);

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (((map.l_getsector1(l) == selected_floor || map.l_getsector2(l) == selected_floor) && selected_floor != -1) ||
			((map.l_getsector1(l) == selected_ceil || map.l_getsector2(l) == selected_ceil) && selected_ceil != -1))
		{
			rect_t lrect = map.l_getrect(l);

			glBegin(GL_LINES);
			glVertex3f(lrect.x1() * scale, lrect.y1() * scale, height);
			glVertex3f(lrect.x2() * scale, lrect.y2() * scale, height);
			glEnd();
		}
	}

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LINE_SMOOTH);
	glLineWidth(1.0f);
}

// render_ssector: Draws a subsector (floor & ceiling)
// ------------------------------------------------ >>
void render_ssector(DWORD ssect)
{
	int light = 0;

	glCullFace(GL_FRONT);

	// Floor
	if (sects[ssect]->f_tex->tex_id == sky_tex_id && render_sky_fullbright)
		light = 255;
	else
		light = sects[ssect]->light;

	set_light(light);

	glBindTexture(GL_TEXTURE_2D, sects[ssect]->f_tex->tex_id);

	glBegin(GL_POLYGON);
	for (DWORD p = 0; p < sects[ssect]->floor->n_points; p++)
	{
		glTexCoord2f((sects[ssect]->floor->points[p].x / scale) * (1.0f / sects[ssect]->f_tex->width),
						(-sects[ssect]->floor->points[p].y / scale) * (1.0f / sects[ssect]->f_tex->height));

		glVertex3f(sects[ssect]->floor->points[p].x,
					sects[ssect]->floor->points[p].y,
					sects[ssect]->floor->points[p].z);
	}
	glEnd();

	if (sects[ssect]->parent_sector == selected_floor && render_hilight)
	{
		glDisable(GL_TEXTURE_2D);
		set_colour(col_3d_hilight);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);

		glBegin(GL_POLYGON);
		for (DWORD p = 0; p < sects[ssect]->floor->n_points; p++)
		{
			glTexCoord2f((sects[ssect]->floor->points[p].x / scale) * (1.0f / sects[ssect]->f_tex->width),
							(-sects[ssect]->floor->points[p].y / scale) * (1.0f / sects[ssect]->f_tex->height));

			glVertex3f(sects[ssect]->floor->points[p].x,
						sects[ssect]->floor->points[p].y,
						sects[ssect]->floor->points[p].z);
		}
		glEnd();

		glEnable(GL_TEXTURE_2D);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	glCullFace(GL_BACK);

	// Ceiling
	if (sects[ssect]->c_tex->tex_id == sky_tex_id && render_sky_fullbright)
		light = 255;
	else
		light = sects[ssect]->light;

	set_light(light);

	glBindTexture(GL_TEXTURE_2D, sects[ssect]->c_tex->tex_id);

	glBegin(GL_POLYGON);
	for (p = 0; p < sects[ssect]->ceil->n_points; p++)
	{
		glTexCoord2f((sects[ssect]->ceil->points[p].x / scale) * (1.0f / sects[ssect]->c_tex->width),
						(-sects[ssect]->ceil->points[p].y / scale) * (1.0f / sects[ssect]->c_tex->height));

		glVertex3f(sects[ssect]->ceil->points[p].x,
					sects[ssect]->ceil->points[p].y,
					sects[ssect]->ceil->points[p].z);
	}
	glEnd();

	if (sects[ssect]->parent_sector == selected_ceil && render_hilight)
	{
		glDisable(GL_TEXTURE_2D);
		set_colour(col_3d_hilight);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);

		glBegin(GL_POLYGON);
		for (p = 0; p < sects[ssect]->ceil->n_points; p++)
		{
			glTexCoord2f((sects[ssect]->ceil->points[p].x / scale) * (1.0f / sects[ssect]->c_tex->width),
							(-sects[ssect]->ceil->points[p].y / scale) * (1.0f / sects[ssect]->c_tex->height));

			glVertex3f(sects[ssect]->ceil->points[p].x,
						sects[ssect]->ceil->points[p].y,
						sects[ssect]->ceil->points[p].z);
		}
		glEnd();

		glEnable(GL_TEXTURE_2D);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

// render_flats: Draws any visible flats
// ---------------------------------- >>
void render_flats()
{
	for (DWORD s = 0; s < visible_subsectors.n_numbers; s++)
			render_ssector(visible_subsectors.numbers[s]);

	if (render_ssect_lines)
	{
		glDisable(GL_TEXTURE_2D);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glLineWidth(3.0f);

		for (DWORD s = 0; s < visible_subsectors.n_numbers; s++)
			render_ssector(visible_subsectors.numbers[s]);

		glEnable(GL_TEXTURE_2D);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glLineWidth(1.0f);
	}

	if (render_hilight)
		render_flat_hilight();
}

// render_wallrect: Draws a 3d rectangle
// ---------------------------------- >>
void render_wallrect(rect_3d_t *rect, short light)
{
	if (light != -1)
	{
		// Set light level
		if (rect->tex == sky_tex_id)
			set_light(255);
		else
			set_light(light);

		// Bind texture
		glBindTexture(GL_TEXTURE_2D, rect->tex);

		// Draw
		glBegin(GL_QUADS);
		glTexCoord2f(rect->tex_coords.x1, rect->tex_coords.y2);
		glVertex3f(rect->tl.x, rect->tl.y, rect->br.z);
		glTexCoord2f(rect->tex_coords.x2, rect->tex_coords.y2);
		glVertex3f(rect->br.x, rect->br.y, rect->br.z);
		glTexCoord2f(rect->tex_coords.x2, rect->tex_coords.y1);
		glVertex3f(rect->br.x, rect->br.y, rect->tl.z);
		glTexCoord2f(rect->tex_coords.x1, rect->tex_coords.y1);
		glVertex3f(rect->tl.x, rect->tl.y, rect->tl.z);
		glEnd();
	}

	if (rect == selected_rect && render_hilight)
	{
		glDisable(GL_TEXTURE_2D);
		set_colour(col_3d_hilight);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);

		glBegin(GL_QUADS);
		glVertex3f(rect->tl.x, rect->tl.y, rect->br.z);
		glVertex3f(rect->br.x, rect->br.y, rect->br.z);
		glVertex3f(rect->br.x, rect->br.y, rect->tl.z);
		glVertex3f(rect->tl.x, rect->tl.y, rect->tl.z);
		glVertex3f(rect->tl.x, rect->tl.y, rect->br.z);
		glEnd();

		glDisable(GL_DEPTH_TEST);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glEnable(GL_LINE_SMOOTH);
		glLineWidth(2.0f);

		set_colour(col_3d_hilight_line);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glBegin(GL_QUADS);
		glVertex3f(rect->tl.x, rect->tl.y, rect->br.z);
		glVertex3f(rect->br.x, rect->br.y, rect->br.z);
		glVertex3f(rect->br.x, rect->br.y, rect->tl.z);
		glVertex3f(rect->tl.x, rect->tl.y, rect->tl.z);
		glVertex3f(rect->tl.x, rect->tl.y, rect->br.z);
		glEnd();

		glEnable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glDisable(GL_LINE_SMOOTH);
		glLineWidth(1.0f);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

// render_wall: Draws a wall
// ---------------------- >>
void render_wall(DWORD l, bool mask)
{
	short light = 0;

	// << --- First Side --- >>
	if (map.lines[l]->side1 != -1)
		light = map.sectors[map.l_getsector1(l)]->light;

	// Draw rects
	if (walls[l].has_lo[0] && !mask)
		render_wallrect(&walls[l].lo[0], light);

	if (walls[l].has_up[0] && !mask)
		render_wallrect(&walls[l].up[0], light);

	if (walls[l].has_mid[0] && mask == walls[l].mask[0])
		render_wallrect(&walls[l].mid[0], light);

	// << --- Second Side --- >>

	// Set light level
	if (map.lines[l]->side2 != -1)
		light = map.sectors[map.l_getsector2(l)]->light;

	// Draw rects
	if (walls[l].has_lo[1] && !mask)
		render_wallrect(&walls[l].lo[1], light);

	if (walls[l].has_up[1] && !mask)
		render_wallrect(&walls[l].up[1], light);

	if (walls[l].has_mid[1] && mask == walls[l].mask[1])
		render_wallrect(&walls[l].mid[1], light);
}

// render_walls: Draws any currently visible map walls
// ------------------------------------------------ >>
void render_walls(bool mask)
{
	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (walls[l].visible)
			render_wall(l, mask);
	}

	if (selected_rect && render_hilight)
		render_wallrect(selected_rect, -1);
}

// render_sprites: Draws any currently visible things
// ----------------------------------------------- >>
void render_sprites()
{
	for (int s = sprite_order.n_numbers - 1; s >= 0; s--)
	{
		int t = sprite_order.numbers[s];
		rect_3d_t rect;
		point3_t p_vector;
		p_vector = cross(camera.view - camera.position, camera.up_vector);
		p_vector = p_vector.normalize();
		int f_height = 0;

		if (sprites3d[t].sector < map.n_sectors)
		{
			if (sprites3d[t].hanging)
				f_height = map.sectors[sprites3d[t].sector]->c_height - sprites3d[t].height;
			else
				f_height = map.sectors[sprites3d[t].sector]->f_height;
		}

		if (map.things[t]->z != 0 && map.hexen)
			f_height = map.things[t]->z;

		if (sprites3d[t].width == -1)
		{
			rect.tl.set((map.things[t]->x - p_vector.x * 8) * scale,
						(map.things[t]->y - p_vector.y * 8) * scale,
						(f_height + 16) * scale);
			rect.br.set((map.things[t]->x + p_vector.x * 8) * scale,
						(map.things[t]->y + p_vector.y * 8) * scale,
						f_height * scale);
		}
		else
		{
			rect.tl.set((map.things[t]->x - p_vector.x * (sprites3d[t].width / 2)) * scale,
						(map.things[t]->y - p_vector.y * (sprites3d[t].width / 2)) * scale,
						(f_height + sprites3d[t].height) * scale);
			rect.br.set((map.things[t]->x + p_vector.x * (sprites3d[t].width / 2)) * scale,
						(map.things[t]->y + p_vector.y * (sprites3d[t].width / 2)) * scale,
						f_height * scale);
		}

		int light = 255;
		rgba_t col = sprites3d[t].colour;

		if (sprites3d[t].sector < map.n_sectors && !render_fullbright)
			light = map.sectors[sprites3d[t].sector]->light;
		else
			light = 255;

		glColor3f((float(col.r / 255.0f) * light) / 255.0f,
					(float(col.g / 255.0f) * light) / 255.0f,
					(float(col.b / 255.0f) * light) / 255.0f);
		glFogf(GL_FOG_END, ((light/255.0f) * 30.0f) + 1.0f);

		/*
		if (map.things[t]->draw_angle)
		{
			glDisable(GL_TEXTURE_2D);
			glBegin(GL_LINES);
			glVertex3f(map.things[t]->x * scale, map.things[t]->y * scale, (f_height + sprites3d[t].height / 2) * scale);
			glVertex3f((map.things[t]->x + 16) * scale, (map.things[t]->y + 16) * scale, (f_height + sprites3d[t].height / 2) * scale);
			glEnd();
			glEnable(GL_TEXTURE_2D);
		}
		*/

		// Bind texture
		glBindTexture(GL_TEXTURE_2D, sprites3d[t].tex_id);

		// Draw
		glBegin(GL_QUADS);
		glTexCoord2f(sprites3d[t].tex_coords.x1, sprites3d[t].tex_coords.y2);
		glVertex3f(rect.tl.x, rect.tl.y, rect.br.z);
		glTexCoord2f(sprites3d[t].tex_coords.x2, sprites3d[t].tex_coords.y2);
		glVertex3f(rect.br.x, rect.br.y, rect.br.z);
		glTexCoord2f(sprites3d[t].tex_coords.x2, sprites3d[t].tex_coords.y1);
		glVertex3f(rect.br.x, rect.br.y, rect.tl.z);
		glTexCoord2f(sprites3d[t].tex_coords.x1, sprites3d[t].tex_coords.y1);
		glVertex3f(rect.tl.x, rect.tl.y, rect.tl.z);
		glEnd();
	}
}

// setup_ssect: Sets up a subsector
// ----------------------------- >>
void setup_ssect(DWORD ssect)
{
	sects[ssect] = new sectinfo_t;

	int floor = 0;
	int ceil = 0;

	int startseg = gl_ssectors[ssect].startseg;
	int endseg = startseg + gl_ssectors[ssect].n_segs;

	// Get textures & heights
	for (DWORD s = startseg; s < endseg; s++)
	{
		// If the seg runs along a line
		if (gl_segs[s].line != SEG_MINISEG)
		{
			int sector = 0;

			if (gl_segs[s].flags & SEG_FLAGS_SIDE)
				sector = map.l_getsector2(gl_segs[s].line);
			else
				sector = map.l_getsector1(gl_segs[s].line);

			if (map.sectors[sector]->f_tex == "F_SKY1")
				sects[ssect]->f_tex = tex_list.get_tex(skytex);
			else
				sects[ssect]->f_tex = tex_list.get_tex(map.sectors[sector]->f_tex, true, true);
			
			if (map.sectors[sector]->c_tex == "F_SKY1")
				sects[ssect]->c_tex = tex_list.get_tex(skytex);
			else
				sects[ssect]->c_tex = tex_list.get_tex(map.sectors[sector]->c_tex, true, true);

			sects[ssect]->light = map.sectors[sector]->light;

			floor = map.sectors[sector]->f_height;
			ceil = map.sectors[sector]->c_height;
			sects[ssect]->parent_sector = sector;
			break;
		}
	}

	// Setup polys

	// Add the first vertex to the floor/ceil polys
	int v = gl_segs[startseg].vertex1;

	if (v & SEG_GLVERTEX)
	{
		sects[ssect]->floor->add_point(point3_t(gl_verts[v & ~SEG_GLVERTEX].x * scale, 
												gl_verts[v & ~SEG_GLVERTEX].y * scale, floor * scale));
		sects[ssect]->ceil->add_point(point3_t(gl_verts[v & ~SEG_GLVERTEX].x * scale, 
												gl_verts[v & ~SEG_GLVERTEX].y * scale, ceil * scale));
	}
	else
	{
		sects[ssect]->floor->add_point(point3_t(map.verts[v]->x * scale, map.verts[v]->y * scale, floor * scale));
		sects[ssect]->ceil->add_point(point3_t(map.verts[v]->x * scale, map.verts[v]->y * scale, ceil * scale));
	}

	// Add each seg's second vertex to the floor/ceil polys
	for (s = startseg; s < endseg; s++)
	{
		v = gl_segs[s].vertex2;

		if (v & SEG_GLVERTEX)
		{
			sects[ssect]->floor->add_point(point3_t(gl_verts[v & ~SEG_GLVERTEX].x * scale, 
													gl_verts[v & ~SEG_GLVERTEX].y * scale, floor * scale));
			sects[ssect]->ceil->add_point(point3_t(gl_verts[v & ~SEG_GLVERTEX].x * scale, 
													gl_verts[v & ~SEG_GLVERTEX].y * scale, ceil * scale));
		}
		else
		{
			sects[ssect]->floor->add_point(point3_t(map.verts[v]->x * scale, map.verts[v]->y * scale, floor * scale));
			sects[ssect]->ceil->add_point(point3_t(map.verts[v]->x * scale, map.verts[v]->y * scale, ceil * scale));
		}
	}
}

void set_visbuffer(int blocked)
{
	for (int a = 0; a < 3600; a++)
		vis_buffer[a] = blocked;
}

void open_view()
{
	int mlook_mod = (int)(((camera.view.z - camera.position.z) * 600.0f));

	if (mlook_mod < 0)
		mlook_mod = -mlook_mod;

	for (int a = 0; a < 600 + mlook_mod; a++)
		vis_buffer[a] = 0;

	for (a = 2700 - mlook_mod; a < 3600; a++)
		vis_buffer[a] = 0;
}

bool seg_is_visible(float x1, float y1, float x2, float y2)
{
	short a1 = (short)(angle_between_vectors(camera.view.x - camera.position.x,
											camera.view.y - camera.position.y,
											x1 - camera.position.x,
											y1 - camera.position.y) * 10.0f);
	short a2 = (short)(angle_between_vectors(camera.view.x - camera.position.x,
											camera.view.y - camera.position.y,
											x2 - camera.position.x,
											y2 - camera.position.y) * 10.0f);

	if (!determine_line_side(camera.position.x, camera.position.y, camera.view.x, camera.view.y, x1, y1))
		a1 = 3600 - a1;

	if (!determine_line_side(camera.position.x, camera.position.y, camera.view.x, camera.view.y, x2, y2))
		a2 = 3600 - a2;

	if (a1 > a2)
	{
		for (short a = a1; a < 3600; a++)
		{
			if (!vis_buffer[a])
				return true;
		}

		for (a = 0; a < a2; a++)
		{
			if (!vis_buffer[a])
				return true;
		}
	}
	else
	{
		for (short a = a1; a < a2; a++)
		{
			if (!vis_buffer[a])
				return true;
		}
	}

	return false;
}

void block_seg(float x1, float y1, float x2, float y2)
{
	short a1 = (short)(angle_between_vectors(camera.view.x - camera.position.x,
											camera.view.y - camera.position.y,
											x1 - camera.position.x,
											y1 - camera.position.y) * 10.0f);
	short a2 = (short)(angle_between_vectors(camera.view.x - camera.position.x,
											camera.view.y - camera.position.y,
											x2 - camera.position.x,
											y2 - camera.position.y) * 10.0f);

	if (!determine_line_side(camera.position.x, camera.position.y, camera.view.x, camera.view.y, x1, y1))
		a1 = 3600 - a1;

	if (!determine_line_side(camera.position.x, camera.position.y, camera.view.x, camera.view.y, x2, y2))
		a2 = 3600 - a2;

	if (a1 > a2)
	{
		for (short a = a1; a < 3600; a++)
			vis_buffer[a] = 1;

		for (a = 0; a < a2; a++)
			vis_buffer[a] = 1;
	}
	else
	{
		for (short a = a1; a < a2; a++)
			vis_buffer[a] = 1;
	}
}

bool view_buffer_full()
{
	if (memchr(vis_buffer, 0, 3600) == NULL)
		return true;
	else
		return false;
}

void process_subsector(short subsect)
{
	bool visible = false;

	// Cycle through segs
	int start = gl_ssectors[subsect].startseg;
	int end = start + gl_ssectors[subsect].n_segs;

	for (int s = start; s < end; s++)
	{
		// Get x1,y1,x2,y2 of seg
		float x1 = 0.0f;
		float y1 = 0.0f;
		float x2 = 0.0f;
		float y2 = 0.0f;
		long v = gl_segs[s].vertex1;

		if (v & SEG_GLVERTEX)
		{
			x1 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
			y1 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
		}
		else
		{
			x1 = map.verts[v]->x * scale;
			y1 = map.verts[v]->y * scale;
		}

		v = gl_segs[s].vertex2;

		if (v & SEG_GLVERTEX)
		{
			x2 = gl_verts[v & ~SEG_GLVERTEX].x * scale;
			y2 = gl_verts[v & ~SEG_GLVERTEX].y * scale;
		}
		else
		{
			x2 = map.verts[v]->x * scale;
			y2 = map.verts[v]->y * scale;
		}

		rect_t seg;
		seg.set(x1, y1, x2, y2);

		// If seg runs along a line
		if (gl_segs[s].line != -1)
		{
			bool side = true;

			if (gl_segs[s].flags & SEG_FLAGS_SIDE)
				side = false;

			// If we're on the right side of that line
			if (determine_line_side(gl_segs[s].line, camera.position.x / scale, camera.position.y / scale) == side)
			{
				// If the seg isn't blocked
				if (seg_is_visible(x1, y1, x2, y2))
				{
					// Block the seg if the line is 1-sided
					if (map.lines[gl_segs[s].line]->side2 == -1 && side)
					{
						if (render_bsp)
							block_seg(x1, y1, x2, y2);
					}
					else
					{
						// Block the seg if it's floor and ceiling are equal or overlapping
						int sector;

						if (side)
							sector = map.l_getsector2(gl_segs[s].line);
						else
							sector = map.l_getsector1(gl_segs[s].line);

						if (sector != -1)
						{
							if (map.sectors[sector]->f_height >= map.sectors[sector]->c_height)
								block_seg(x1, y1, x2, y2);
						}
					}

					// The line and subsector are visible
					walls[gl_segs[s].line].visible = true;
					visible = true;
				}
			}
		}
		else // Seg doesn't run along a line
		{
			// If we're on the right side of the seg
			if (determine_line_side(seg, camera.position.x, camera.position.y))
			{
				// If the seg isn't blocked, the ssector is visible
				if (seg_is_visible(x1, y1, x2, y2))
					visible = true;
			}
		}
	}

	if (visible)
	{
		visible_subsectors.add(subsect, true);
		//sects[subsect]->visible = true;
	}
}

// *****************************************************************************
// *****************************************************************************
// *****************************************************************************
// *****************************************************************************
void walk_bsp_tree(unsigned short node)
{
	// If we have a subsector
	if (node & CHILD_IS_SUBSEC)
	{
		short subsect = node & ~CHILD_IS_SUBSEC;
		process_subsector(subsect);

		return;
	}

	if (view_buffer_full())
		return;

	int x1 = gl_nodes[node].x;
	int y1 = gl_nodes[node].y;
	int x2 = x1 + gl_nodes[node].dx;
	int y2 = y1 + gl_nodes[node].dy;

	rect_t part(x1, y1, x2, y2);

	if (determine_line_side_f(part, camera.position.x / scale, camera.position.y / scale) > 0)
	{
		walk_bsp_tree(gl_nodes[node].right_child);
		walk_bsp_tree(gl_nodes[node].left_child);
	}
	else if (determine_line_side_f(part, camera.position.x / scale, camera.position.y / scale) < 0)
	{
		walk_bsp_tree(gl_nodes[node].left_child);
		walk_bsp_tree(gl_nodes[node].right_child);
	}
	else
	{
		if (determine_line_side(part, camera.view.x / scale, camera.view.y / scale))
		{
			walk_bsp_tree(gl_nodes[node].right_child);
			walk_bsp_tree(gl_nodes[node].left_child);
		}
		else
		{
			walk_bsp_tree(gl_nodes[node].left_child);
			walk_bsp_tree(gl_nodes[node].right_child);
		}
	}
}

// setup_sects: Sets up map subsectors
// -------------------------------- >>
void setup_sects()
{
	sects = (sectinfo_t **)realloc(sects, n_gl_ssectors * sizeof(sectinfo_t *));

	for (DWORD s = 0; s < n_gl_ssectors; s++)
		setup_ssect(s);
}

// cleanup_3dmode: Cleans up memory used by 3d mode
// --------------------------------------------- >>
void cleanup_3dmode()
{
	if (sects)
	{
		for (DWORD s = 0; s < n_gl_ssectors; s++)
		{
			free(sects[s]->ceil->points);
			free(sects[s]->floor->points);
			free(sects[s]->ceil);
			free(sects[s]->floor);
			free(sects[s]);
		}
	}
}

// setup_wall_rect: Sets up a wall rect
// --------------------------------- >>
void setup_wall_rect(DWORD l, int side, int part)
{
	rect_t r = map.l_getrect(l);

	// Middle
	if (part == 0)
	{
		// First side
		if (side == 1)
		{
			walls[l].mask[0] = false;
			walls[l].has_mid[0] = false;

			// Solid 1s wall
			if (map.lines[l]->side2 == -1)
			{
				walls[l].has_mid[0] = true;
				walls[l].mid[0].tl.set(r.x1() * scale, r.y1() * scale, map.sectors[map.l_getsector1(l)]->c_height * scale);
				walls[l].mid[0].br.set(r.x2() * scale, r.y2() * scale, map.sectors[map.l_getsector1(l)]->f_height * scale);
				selected_rect = &walls[l].mid[0];
			}

			// Mid texture (2s wall)
			if (map.lines[l]->side2 != -1 && map.sides[map.lines[l]->side1]->tex_middle != "-")
			{
				int yoff = map.l_getyoff(l, 1);
				int lower = max(map.sectors[map.l_getsector1(l)]->f_height, map.sectors[map.l_getsector2(l)]->f_height);
				int upper = min(map.sectors[map.l_getsector1(l)]->c_height, map.sectors[map.l_getsector2(l)]->c_height);
				int height = tex_list.get_tex(map.sides[map.lines[l]->side1]->tex_middle)->height;

				walls[l].has_mid[0] = true;
				walls[l].mask[0] = true;
				
				//if (height > (upper - lower))
					//height = upper - lower;

				if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
				{
					int top = lower + height + yoff;
					int bottom = lower + yoff;

					if (bottom < lower)
						bottom = lower;

					if (top > upper)
						top = upper;

					walls[l].mid[0].tl.set(r.x1() * scale, r.y1() * scale, top * scale);
					walls[l].mid[0].br.set(r.x2() * scale, r.y2() * scale, bottom * scale);
				}
				else
				{
					int top = upper + yoff;
					int bottom = top - height/* + yoff*/;

					if (bottom < lower)
						bottom = lower;

					if (top > upper)
						top = upper;

					walls[l].mid[0].tl.set(r.x1() * scale, r.y1() * scale, top * scale);
					walls[l].mid[0].br.set(r.x2() * scale, r.y2() * scale, bottom * scale);
				}
			}
		}

		// Second side
		if (side == 2)
		{
			walls[l].mask[1] = false;
			walls[l].has_mid[1] = false;

			// Mid texture (2s wall)
			if (map.sides[map.lines[l]->side2]->tex_middle != "-")
			{
				int yoff = map.l_getyoff(l, 2);
				int lower = max(map.sectors[map.l_getsector1(l)]->f_height, map.sectors[map.l_getsector2(l)]->f_height);
				int upper = min(map.sectors[map.l_getsector1(l)]->c_height, map.sectors[map.l_getsector2(l)]->c_height);
				int height = tex_list.get_tex(map.sides[map.lines[l]->side2]->tex_middle)->height;

				walls[l].has_mid[1] = true;
				walls[l].mask[1] = true;
				
				/*
				if (height > (upper - lower))
					height = upper - lower;
					*/

				if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
				{
					int top = (lower + height) + yoff;
					int bottom = lower + yoff;

					if (bottom < lower)
						bottom = lower;

					if (top > upper)
						top = upper;

					walls[l].mid[1].tl.set(r.x2() * scale, r.y2() * scale, top * scale);
					walls[l].mid[1].br.set(r.x1() * scale, r.y1() * scale, bottom * scale);
				}
				else
				{
					int top = upper + yoff;
					int bottom = top - height/* + yoff*/;

					if (bottom < lower)
						bottom = lower;

					if (top > upper)
						top = upper;

					walls[l].mid[1].tl.set(r.x2() * scale, r.y2() * scale, top * scale);
					walls[l].mid[1].br.set(r.x1() * scale, r.y1() * scale, bottom * scale);
				}
			}
		}
	}

	// Upper
	if (part == 1)
	{
		// First side
		if (side == 1)
		{
			walls[l].has_up[0] = true;
			walls[l].up[0].tl.set(r.x1() * scale, r.y1() * scale, map.sectors[map.l_getsector1(l)]->c_height * scale);
			walls[l].up[0].br.set(r.x2() * scale, r.y2() * scale, map.sectors[map.l_getsector2(l)]->c_height * scale);
		}

		// Second side
		if (side == 2)
		{
			walls[l].has_up[1] = true;
			walls[l].up[1].tl.set(r.x2() * scale, r.y2() * scale, map.sectors[map.l_getsector2(l)]->c_height * scale);
			walls[l].up[1].br.set(r.x1() * scale, r.y1() * scale, map.sectors[map.l_getsector1(l)]->c_height * scale);
		}
	}

	// Lower
	if (part == 2)
	{
		// First side
		if (side == 1)
		{
			walls[l].has_lo[0] = true;
			walls[l].lo[0].tl.set(r.x1() * scale, r.y1() * scale, map.sectors[map.l_getsector2(l)]->f_height * scale);
			walls[l].lo[0].br.set(r.x2() * scale, r.y2() * scale, map.sectors[map.l_getsector1(l)]->f_height * scale);
		}

		// Second side
		if (side == 2)
		{
			walls[l].has_lo[1] = true;
			walls[l].lo[1].tl.set(r.x2() * scale, r.y2() * scale, map.sectors[map.l_getsector1(l)]->f_height * scale);
			walls[l].lo[1].br.set(r.x1() * scale, r.y1() * scale, map.sectors[map.l_getsector2(l)]->f_height * scale);
		}
	}
}

// setup_wall_tex: Sets up a wall texture coordinates
// ----------------------------------------------- >>
void setup_wall_tex(DWORD l, int side, int part)
{
	texture_t	*tex;
	int			x1=0, y1=0, x2=0, y2=0;
	rect_t		r = map.l_getrect(l);
	int			length = distance(r.x1(), r.y1(), r.x2(), r.y2());
	int			height = 0;

	if (part == 0)
	{
		if (side == 1)
		{
			height = map.sectors[map.l_getsector1(l)]->c_height - map.sectors[map.l_getsector1(l)]->f_height;

			// Get texture
			tex = tex_list.get_tex(map.sides[map.lines[l]->side1]->tex_middle, true, false);

			// Set tex id
			walls[l].mid[0].tex = tex->tex_id;

			// Setup tex coords
			if (map.lines[l]->side2 == -1)
			{
				// 1s wall
				if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
				{
					x1 = map.sides[map.lines[l]->side1]->x_offset;
					x2 = x1 + length;
					y2 = map.sides[map.lines[l]->side1]->y_offset;
					y1 = y2 - height;
				}
				else
				{
					x1 = map.sides[map.lines[l]->side1]->x_offset;
					x2 = x1 + length;
					y1 = map.sides[map.lines[l]->side1]->y_offset;
					y2 = y1 + height;
				}
			}
			else if (walls[l].has_mid[0])
			{
				// 2s wall
				int lower = max(map.sectors[map.l_getsector1(l)]->f_height, map.sectors[map.l_getsector2(l)]->f_height);
				int upper = min(map.sectors[map.l_getsector1(l)]->c_height, map.sectors[map.l_getsector2(l)]->c_height);
				int height = upper - lower;

				int rect_upper = 0;
				int rect_lower = 0;

				if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
				{
					rect_lower = lower + map.sides[map.lines[l]->side1]->y_offset;
					rect_upper = rect_lower + tex->height;
				}
				else
				{
					rect_upper = upper + map.sides[map.lines[l]->side1]->y_offset;
					rect_lower = rect_upper - tex->height;
				}

				int rect_height = min(rect_upper, upper) - max(rect_lower, lower);

				x1 = map.sides[map.lines[l]->side1]->x_offset;
				x2 = x1 + length;
				y1 = 0;
				y2 = y1 + rect_height;
				
				if (rect_upper > upper)
				{
					y1 = rect_upper - upper;
					y2 = y1 + rect_height;
				}
			}

			walls[l].mid[0].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
			walls[l].mid[0].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
			walls[l].mid[0].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
			walls[l].mid[0].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
		}

		if (side == 2)
		{
			height = map.sectors[map.l_getsector2(l)]->c_height - map.sectors[map.l_getsector2(l)]->f_height;

			// Get texture
			tex = tex_list.get_tex(map.sides[map.lines[l]->side2]->tex_middle, true, false);

			// Set tex id
			walls[l].mid[1].tex = tex->tex_id;

			// 2s wall
			if (walls[l].has_mid[1])
			{
				int lower = max(map.sectors[map.l_getsector1(l)]->f_height, map.sectors[map.l_getsector2(l)]->f_height);
				int upper = min(map.sectors[map.l_getsector1(l)]->c_height, map.sectors[map.l_getsector2(l)]->c_height);
				int height = upper - lower;

				int rect_upper = 0;
				int rect_lower = 0;

				if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
				{
					rect_lower = lower + map.sides[map.lines[l]->side2]->y_offset;
					rect_upper = rect_lower + tex->height;
				}
				else
				{
					rect_upper = upper + map.sides[map.lines[l]->side2]->y_offset;
					rect_lower = rect_upper - tex->height;
				}

				int rect_height = min(rect_upper, upper) - max(rect_lower, lower);

				x1 = map.sides[map.lines[l]->side2]->x_offset;
				x2 = x1 + length;
				y1 = 0;
				y2 = y1 + rect_height;
				
				if (rect_upper > upper)
				{
					y1 = rect_upper - upper;
					y2 = y1 + rect_height;
				}

				walls[l].mid[1].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
				walls[l].mid[1].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
				walls[l].mid[1].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
				walls[l].mid[1].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
			}
		}
	}

	if (part == 1)
	{
		if (side == 1)
		{
			height = map.sectors[map.l_getsector1(l)]->c_height - map.sectors[map.l_getsector2(l)]->c_height;

			if (map.sectors[map.l_getsector1(l)]->c_tex == "F_SKY1"
				&& map.sectors[map.l_getsector2(l)]->c_tex == "F_SKY1")
				tex = tex_list.get_tex(skytex);
			else
				tex = tex_list.get_tex(map.sides[map.lines[l]->side1]->tex_upper, true, false);

			walls[l].up[0].tex = tex->tex_id;

			// Setup tex coords
			if (map.lines[l]->flags & LINE_UPPERUNPEGGED)
			{
				x1 = map.sides[map.lines[l]->side1]->x_offset;
				x2 = x1 + length;
				y1 = map.sides[map.lines[l]->side1]->y_offset;
				y2 = y1 + height;
			}
			else
			{
				x1 = map.sides[map.lines[l]->side1]->x_offset;
				x2 = x1 + length;
				y2 = tex->height + map.sides[map.lines[l]->side1]->y_offset;
				y1 = y2 - height;
			}

			walls[l].up[0].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
			walls[l].up[0].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
			walls[l].up[0].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
			walls[l].up[0].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
		}

		if (side == 2)
		{
			height = map.sectors[map.l_getsector2(l)]->c_height - map.sectors[map.l_getsector1(l)]->c_height;

			if (map.sectors[map.l_getsector1(l)]->c_tex == "F_SKY1"
				&& map.sectors[map.l_getsector2(l)]->c_tex == "F_SKY1")
				tex = tex_list.get_tex(skytex);
			else
				tex = tex_list.get_tex(map.sides[map.lines[l]->side2]->tex_upper, true, false);

			walls[l].up[1].tex = tex->tex_id;

			// Setup tex coords
			if (map.lines[l]->flags & LINE_UPPERUNPEGGED)
			{
				x1 = map.sides[map.lines[l]->side2]->x_offset;
				x2 = x1 + length;
				y1 = map.sides[map.lines[l]->side2]->y_offset;
				y2 = y1 + height;
			}
			else
			{
				x1 = map.sides[map.lines[l]->side2]->x_offset;
				x2 = x1 + length;
				y2 = tex->height + map.sides[map.lines[l]->side2]->y_offset;
				y1 = y2 - height;
			}

			walls[l].up[1].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
			walls[l].up[1].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
			walls[l].up[1].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
			walls[l].up[1].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
		}
	}

	if (part == 2)
	{
		if (side == 1)
		{
			height = map.sectors[map.l_getsector2(l)]->f_height - map.sectors[map.l_getsector1(l)]->f_height;
			tex = tex_list.get_tex(map.sides[map.lines[l]->side1]->tex_lower, true, false);
			walls[l].lo[0].tex = tex->tex_id;

			int height2 = map.sectors[map.l_getsector1(l)]->c_height
							- map.sectors[map.l_getsector2(l)]->f_height;

			// Setup tex coords
			if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
			{
				x1 = map.sides[map.lines[l]->side1]->x_offset;
				x2 = x1 + length;
				y1 = height2 + map.sides[map.lines[l]->side1]->y_offset;
				y2 = y1 + height;
			}
			else
			{
				x1 = map.sides[map.lines[l]->side1]->x_offset;
				x2 = x1 + length;
				y1 = map.sides[map.lines[l]->side1]->y_offset;
				y2 = y1 + height;
			}

			walls[l].lo[0].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
			walls[l].lo[0].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
			walls[l].lo[0].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
			walls[l].lo[0].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
		}

		if (side == 2)
		{
			height = map.sectors[map.l_getsector1(l)]->f_height - map.sectors[map.l_getsector2(l)]->f_height;
			tex = tex_list.get_tex(map.sides[map.lines[l]->side2]->tex_lower, true, false);
			walls[l].lo[1].tex = tex->tex_id;

			int height2 = map.sectors[map.l_getsector2(l)]->c_height
							- map.sectors[map.l_getsector1(l)]->f_height;

			// Setup tex coords
			if (map.lines[l]->flags & LINE_LOWERUNPEGGED)
			{
				x1 = map.sides[map.lines[l]->side2]->x_offset;
				x2 = x1 + length;
				y1 = height2 + map.sides[map.lines[l]->side2]->y_offset;
				y2 = y1 + height;
			}
			else
			{
				x1 = map.sides[map.lines[l]->side2]->x_offset;
				x2 = x1 + length;
				y1 = map.sides[map.lines[l]->side2]->y_offset;
				y2 = y1 + height;
			}

			walls[l].lo[1].tex_coords.x1 = ((float)x1 / tex->width) * tex->tex_x;
			walls[l].lo[1].tex_coords.y1 = ((float)y1 / tex->height) * tex->tex_y;
			walls[l].lo[1].tex_coords.x2 = ((float)x2 / tex->width) * tex->tex_x;
			walls[l].lo[1].tex_coords.y2 = ((float)y2 / tex->height) * tex->tex_y;
		}
	}
}

// setup_walls: Sets up map walls (rects & textures)
// ---------------------------------------------- >>
void setup_walls()
{
	free(walls);
	walls = NULL;
	walls = (wallinfo_t *)realloc(walls, map.n_lines * sizeof(wallinfo_t));

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		walls[l].has_mid[0] = false;
		walls[l].has_mid[1] = false;
		walls[l].has_up[0] = false;
		walls[l].has_up[1] = false;
		walls[l].has_lo[0] = false;
		walls[l].has_lo[1] = false;

		if (map.lines[l]->side1 != -1)
		{
			if (map.l_needsmidtex(l, 1))
			{
				setup_wall_rect(l, 1, 0);
				setup_wall_tex(l, 1, 0);
			}

			if (map.l_needsuptex(l, 1))
			{
				setup_wall_rect(l, 1, 1);
				setup_wall_tex(l, 1, 1);
			}

			if (map.l_needslotex(l, 1))
			{
				setup_wall_rect(l, 1, 2);
				setup_wall_tex(l, 1, 2);
			}
		}

		if (map.lines[l]->side2 != -1)
		{
			if (map.l_needsmidtex(l, 2))
			{
				setup_wall_rect(l, 2, 0);
				setup_wall_tex(l, 2, 0);
			}

			if (map.l_needsuptex(l, 2))
			{
				setup_wall_rect(l, 2, 1);
				setup_wall_tex(l, 2, 1);
			}

			if (map.l_needslotex(l, 2))
			{
				setup_wall_rect(l, 2, 2);
				setup_wall_tex(l, 2, 2);
			}
		}
	}
}

void setup_sprites()
{
	free(sprites3d);
	sprites3d = NULL;
	sprites3d = (sprite_3d_t *)realloc(sprites3d, map.n_things * sizeof(sprite_3d_t));

	for (DWORD t = 0; t < map.n_things; t++)
	{
		thingdef_t *tdef = get_tdef(map.things[t]->type);
		//texture_t *tex = tex_list.get_tex(tdef->sprite, false, false);
		texture_t *tex = tdef->sprite;

		if (tex)
		{
			if (tex->name[0] != '_')
			{
				sprites3d[t].width = tex->width;
				sprites3d[t].height = tex->height;
				sprites3d[t].colour.set(255, 255, 255, 255);
			}
			else
			{
				sprites3d[t].width = -1;
				sprites3d[t].height = -1;
				sprites3d[t].colour.set(tdef->colour);
			}

			sprites3d[t].tex_id = tex->tex_id;
			sprites3d[t].tex_coords.x1 = 0.0f;
			sprites3d[t].tex_coords.y1 = 0.0f;
			sprites3d[t].tex_coords.x2 = tex->tex_x;
			sprites3d[t].tex_coords.y2 = tex->tex_y;
			sprites3d[t].hanging = tdef->hanging;
		}

		sprites3d[t].sector = determine_sector(map.things[t]->x, map.things[t]->y);
	}
}

void setup_slopes()
{
	// eventually...
}

void setup_3dmode()
{
	sky_tex_id = tex_list.gl_tex(skytex);
	gui_drawsplash("Setting up walls...");
	setup_walls();
	gui_drawsplash("Setting up floors...");
	setup_sects();
	gui_drawsplash("Setting up things...");
	setup_sprites();

	/*
	if (map.zdoom)
	{
		gui_drawsplash("Setting up slopes...");
		setup_slopes();
	}
	*/
}

void clear_visibility()
{
	//for (DWORD s = 0; s < n_gl_ssectors; s++)
		//sects[s]->visible = false;
	visible_subsectors.clear();

	for (DWORD l = 0; l < map.n_lines; l++)
		walls[l].visible = false;
}

void sort_distances(t_dist** a, int length)
{
	for (int p = 1; p <= length; p++)
	{
		t_dist* tmp = a[p];
		int j;

		for (j = p; j > 0 && tmp->distance < a[j - 1]->distance; j--)
			a[j] = a[j - 1];

		a[j] = tmp;
	}
}

void setup_sprite_order()
{
	t_dist** distances = (t_dist **)malloc(map.n_things * sizeof(t_dist *));

	for (DWORD t = 0; t < map.n_things; t++)
	{
		distances[t] = new t_dist;
		distances[t]->thing = t;
		distances[t]->distance = distance(camera.position.x / scale, camera.position.y / scale, map.things[t]->x, map.things[t]->y);
	}

	sort_distances(distances, map.n_things - 1);
	sprite_order.clear();

	for (t = 0; t < map.n_things; t++)
	{
		sprite_order.add(distances[t]->thing, false);
		free(distances[t]);
	}

	free(distances);
}

// render_3dmode: Draws 3d mode
// ------------------------- >>
void render_3dmode()
{
	if (browser)
	{
		// Set 2d screen
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0f, vid_width, vid_height, 0.0f, -1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		draw_browser();
		SDL_GL_SwapBuffers();
		return;
	}

	// Set 3d screen
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, (GLfloat)320/(GLfloat)200, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Clear screen
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(1.0f, 1.0f, 1.0f);

	if (gravity)
		apply_gravity();

	// Set view
	camera.Look();

	// Determine visibility
	ltest = 255;
	clear_visibility();
	set_visbuffer(1);
	open_view();
	walk_bsp_tree(n_gl_nodes - 1);
	//walk_bsp_tree(0, 0);

	// Render
	render_walls(false);
	render_flats();
	render_walls(true);

	if (render_things)
	{
		setup_sprite_order();
		render_sprites();
	}

	// Set 2d screen (for hud)
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f, vid_width, vid_height, 0.0f, -1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	draw_texture("_xhair", (vid_width / 2) - 8, (vid_height / 2) - 8, 16, 16, col_3d_crosshair);

	if (msg_count > 0)
	{
		msg_count--;

		if (msg_count > 20)
			draw_text(0, 0, COL_WHITE, 0, msg_text.c_str());
		else
			draw_text(0, 0, rgba_t(255, 255, 255, ((float)msg_count / 20.0f) * 255), 0, msg_text.c_str());
	}

	if (render_debuginfo)
	{
		point3_t view = camera.view - camera.position;
		draw_text(vid_width, 10, COL_RED, 2, "View Vector: %1.3f, %1.3f, %1.3f", view.x, view.y, view.z);

		int x_int = (int)(camera.position.x / scale);
		int y_int = (int)(camera.position.y / scale);
		int z_int = (int)(camera.position.z / scale);
		draw_text(vid_width, 20, COL_RED, 2, "Position: %d, %d, %d", x_int, y_int, z_int);
		draw_text(vid_width, 30, COL_RED, 2, "Hilighted line: %d", h_line);

		draw_text(vid_width, 40, COL_RED, 2, "Camera Pos: %1.3f, %1.3f, %1.3f", camera.position.x, camera.position.y, camera.position.z);
		draw_text(vid_width, 50, COL_RED, 2, "Camera View: %1.3f, %1.3f, %1.3f", camera.view.x, camera.view.y, camera.view.z);
	}

	// Draw console if open
	if (console_height != 0.0f)
		draw_console();

	glEnable(GL_TEXTURE_2D);

	// Swap buffers
	SDL_GL_SwapBuffers();
}
