
struct point3_t
{
	float x, y, z;

	point3_t()
	{
		x = y = z = 0.0f;
	}

	point3_t(float X, float Y, float Z)
	{
		x = X;
		y = Y;
		z = Z;
	}

	void set(float X, float Y, float Z)
	{
		x = X;
		y = Y;
		z = Z;
	}

	void set(point3_t p)
	{
		x = p.x;
		y = p.y;
		z = p.z;
	}

	float magnitude()
	{
		return (float)sqrt((x * x) + (y * y) + (z * z));
	}

	point3_t normalize()
	{
		float mag = magnitude();

		return point3_t(x / mag, y / mag, z / mag);
	}

	point3_t operator+(point3_t point)
	{
		return point3_t(point.x + x, point.y + y, point.z + z);
	}

	point3_t operator-(point3_t point)
	{
		return point3_t(x - point.x, y - point.y, z - point.z);
	}
	
	point3_t operator*(float num)
	{
		return point3_t(x * num, y * num, z * num);
	}

	point3_t operator/(float num)
	{
		return point3_t(x / num, y / num, z / num);
	}
};

struct tex_coord_t
{
	float x1, y1, x2, y2;

	tex_coord_t()
	{
		x1 = y1 = 0.0f;
		x2 = y2 = 1.0f;
	}
};

struct rect_3d_t
{
	point3_t	tl, br;
	GLuint		tex;
	tex_coord_t	tex_coords;
};

struct wallinfo_t
{
	rect_3d_t	up[2], lo[2], mid[2];
	bool		has_up[2], has_lo[2], has_mid[2];
	bool		mask[2];
	bool		visible;
};

struct polygon_t
{
	point3_t	*points;
	WORD		n_points;

	polygon_t() { points = NULL; n_points = 0; }

	void add_point(point3_t point)
	{
		n_points++;
		points = (point3_t *)realloc(points, n_points * sizeof(point3_t));
		points[n_points - 1].set(point);
	}
};

struct sectinfo_t
{
	polygon_t	*floor;
	polygon_t	*ceil;
	
	texture_t	*f_tex;
	texture_t	*c_tex;

	BYTE		light;
	int			parent_sector;
	int			parent_node;

	bool		visible;

	sectinfo_t()
	{
		f_tex = c_tex = NULL;
		floor = new polygon_t;
		ceil = new polygon_t;
		visible = false;
	}
};

struct sprite_3d_t
{
	short		width;
	short		height;
	WORD		sector;
	GLuint		tex_id;
	tex_coord_t	tex_coords;
	bool		hanging;
	rgba_t		colour;
};

struct gl_seg
{
	long	vertex1;
	long	vertex2;
	short	line;
	short	flags;
	long	partner;
};

struct gl_ssect
{
	long	n_segs;
	long	startseg;
};

struct gl_vertex
{
	int	x;
	int	y;
};

struct gl_node
{
   short x; // partition line
   short y;
   short dx;
   short dy;
   short right_bbox[4];
   short left_bbox[4];
   unsigned short right_child;
   unsigned short left_child;
};

#define CHILD_IS_SUBSEC (1 << 15)
#define SEG_MINISEG		0xFFFF
#define	SEG_FLAGS_SIDE	0x0001
#define SEG_GLVERTEX	0x40000000